import sys
from typing import Any, Dict, Optional

import cv2
import numpy as np
import math
import numfi
from numpy import array, exp, nditer


from scipy.signal import savgol_filter
from scipy.optimize import curve_fit

from vmbpy import *
import matplotlib.pyplot as plt

opencv_display_format = PixelFormat.Bgr8

def print_preamble():
    print('//////////////////////////////////////')
    print('/// VmbPy Lens Shading Calibration ///')
    print('//////////////////////////////////////\n')


def print_usage():
    print('Usage:')
    print('    python lensshading_calibration.py [camera_id] [/{h|e|w}]')
    print()
    print('Parameters:')
    print('    camera_id   ID of the camera to use (using first camera if not specified)')
    print('    /h          Prints this help')
    print('    /e          Evaluation mode, displays the calculated correction values as a graph')
    print('    /w          Meassure once and write values to camera')
    print()

def parse_args() -> Dict[str, Any]:
    args = sys.argv[1:]
    argc = len(args)

    result: Dict[str, Any] = {}

    if (argc <= 0) or (4 <= argc):
        abort(reason="Invalid number of arguments. Abort.", return_code=2, usage=True)

    for arg in args:
        if arg in ('/h'):
            print_usage()
            sys.exit(0)

        # Examine command parameters
        if arg in ('/e', '/w'):
            if result.get('mode') is None:
                result['mode'] = arg
            else:
                abort(reason="Multiple Commands specified. Abort.", return_code=2, usage=True)

        elif arg in ('--once'):
            if result.get('options') is None:
                result['options'] = [arg] 
            else:
                result['options'].append(arg)

        # Examine camera id
        elif result.get('camera_id') is None:
            result['camera_id'] = arg

        else:
            abort(reason="Invalid arguments. Abort.", return_code=2, usage=True)

    # Apply defaults
    if not result.get('mode'):
        result['mode'] = '/e'

    if not result.get('options'):
        result['options'] = [] 

    return result

def abort(reason: str, return_code: int = 1, usage: bool = False):
    print(reason + '\n')

    if usage:
        print_usage()

    sys.exit(return_code)

def get_camera(camera_id: Optional[str]) -> Camera:
    with VmbSystem.get_instance() as vmb:
        if camera_id:
            try:
                return vmb.get_camera_by_id(camera_id)

            except VmbCameraError:
                abort('Failed to access Camera \'{}\'. Abort.'.format(camera_id))

        else:
            cams = vmb.get_all_cameras()
            if not cams:
                abort('No Cameras accessible. Abort.')

            return cams[0]


def setup_camera(cam: Camera):
    with cam:
        # Try to adjust GeV packet size. This Feature is only available for GigE - Cameras.
        try:
            stream = cam.get_streams()[0]
            stream.GVSPAdjustPacketSize.run()
            while not stream.GVSPAdjustPacketSize.is_done():
                pass

        except (AttributeError, VmbFeatureError):
            pass

        try:
            cam.ExposureAuto.set('Off')
            cam.LensShadingEnable.set(False)

        except (AttributeError, VmbFeatureError):
            pass


def setup_pixel_format(cam: Camera):
    # Query available pixel formats. Prefer color formats over monochrome formats
    cam_formats = cam.get_pixel_formats()
    cam_color_formats = intersect_pixel_formats(cam_formats, COLOR_PIXEL_FORMATS)
    convertible_color_formats = tuple(f for f in cam_color_formats
                                      if opencv_display_format in f.get_convertible_formats())

    cam_mono_formats = intersect_pixel_formats(cam_formats, MONO_PIXEL_FORMATS)
    convertible_mono_formats = tuple(f for f in cam_mono_formats
                                     if opencv_display_format in f.get_convertible_formats())

    # if OpenCV compatible color format is supported directly, use that
    if opencv_display_format in cam_formats:
        cam.set_pixel_format(opencv_display_format)

    # else if existing color format can be converted to OpenCV format do that
    elif convertible_color_formats:
        cam.set_pixel_format(convertible_color_formats[0])

    # fall back to a mono format that can be converted
    elif convertible_mono_formats:
        cam.set_pixel_format(convertible_mono_formats[0])

    else:
        abort('Camera does not support an OpenCV compatible format. Abort.')

def integers_to_bytes(integers : list[int], bytes_per_value : int) -> list[int]:
    """
    Convert integer or list of integers to their byte representation

    eg.
    bytes_per_value = 2
    4095 -> [0xff, 0x0f]

    self.assertEqual(self.integers_to_bytes(4095, 2), [0xff, 0x0f])
    self.assertEqual(self.integers_to_bytes([4095, 4094], 2), [0xff, 0x0f, 0xfe, 0x0f])
    """
    # handle single value case - convert to list
    if not isinstance(integers, list):
        integers = [integers]

    result = []
    maximum_possible_value = pow(2,8*bytes_per_value)-1
    for integer in integers:
        assert integer <= maximum_possible_value, f"cannot fit {integer} into {bytes_per_value} bytes"
        assert integer >= 0, "value has to be greater or equal to zero"
        result.extend(list(integer.to_bytes(length=bytes_per_value, byteorder='little', signed=False)))
    return result

def main():
    print_preamble()
    args = parse_args()

    with VmbSystem.get_instance():
        with get_camera(args.get('camera_id')) as cam:
            setup_camera(cam)
            setup_pixel_format(cam)

            mode = args['mode']
            options = args['options']

            fig, ax = plt.subplots()

            if '--once' in options:
                print("Performing exposure once")
                try:
                    cam.ExposureAuto.set('Once')
                except (AttributeError, VmbFeatureError):
                    pass
                for frame in cam.get_frame_generator(limit=None, timeout_ms=2000):
                    exp = str(cam.ExposureAuto.get())
                    if exp == 'Off': 
                        break

            # Acquire 10 frame with a custom timeout (default is 2000ms) per frame acquisition.
            for frame in cam.get_frame_generator(limit=None, timeout_ms=2000):
                print('Got {}'.format(frame), flush=True)
                key = cv2.waitKey(1)

                print('{} acquired {}'.format(cam, frame), flush=True)
                # Convert frame if it is not already the correct format
                if frame.get_pixel_format() == opencv_display_format:
                    display = frame
                else:
                    # This creates a copy of the frame. The original `frame` object can be requeued
                    # safely while `display` is used
                    display = frame.convert_pixel_format(opencv_display_format)

                src = display.as_numpy_ndarray()

                # Convert to grayscale img
                detected = cv2.cvtColor(src, cv2.COLOR_BGR2GRAY)

                # Threshold image
                ret,detectedThres= cv2.threshold(detected,10,255,cv2.THRESH_BINARY)

                # Find center point
                mass_x, mass_y = np.where(detectedThres >= 1)
                cent_x = np.average(mass_x)
                cent_y = np.average(mass_y)
                center = (0,0)
                if not (math.isnan(cent_x) or math.isnan(cent_y)):
                    center = (int(cent_y),int(cent_x))

                # Calculate distance from center point to the furthest corner
                dist1 = np.sqrt(center[0]**2+center[1]**2)
                dist2 = np.sqrt((src.shape[1] - center[0])**2+center[1]**2)
                dist3 = np.sqrt((src.shape[1] - center[0])**2+(src.shape[0] - center[1])**2)
                dist4 = np.sqrt(center[0]**2+(src.shape[0] - center[1])**2)
                maxDistance = int(max(dist1, dist2, dist3, dist4))

                # Calculate the mean image intensity for the given distance from the center point
                imagelineValues = []
                for dist in range(0, maxDistance):
                    #Accumulator used for averaging the circles
                    acc = []

                    # Calculate pixel coordinates given angle and distance from center point
                    for angle in range(0, 360, 1):
                        x = center[0] + math.sin(angle) * dist
                        y = center[1] + math.cos(angle) * dist

                        # Check if pixel is in image
                        if (y >= 0 and y <src.shape[0]) and (x >= 0 and x <src.shape[1]):
                            # Add pixel value to accumulator
                            acc.append(src[int(y),int(x)][0])

                    # Calculate and add average to imagelineValues
                    if len(acc) > 0:
                        avg = sum(acc)/len(acc)
                        imagelineValues.append(avg)

                # Calculate Correction Values
                MAXINTENSITYFACTOR = 1
                maxImagelineValue = max(imagelineValues)
                correctionValues = [min(7.9, maxImagelineValue/(i*MAXINTENSITYFACTOR)) for i in imagelineValues]
                
                #Smoothen Correction Values
                correctionValuesSmooth = savgol_filter(correctionValues, 100, 3)
                correctionValuesSmooth = [min(7.9, i) for i in correctionValuesSmooth]

                # Write Correction Values to Camera, if arg is write
                if mode == '/w':

                    #Convert Values to Dec 16Bit 1Signed, 3Dec, 12Frac
                    lensCorrectionDec = []
                    for i in correctionValuesSmooth:
                        x = numfi(i, 1, 16, 12)
                        y = int(x.base_repr(10)[0])
                        lensCorrectionDec.append(y)
                    print('correctionValue')
                    print(lensCorrectionDec)
                    print('correctionValue end')
                    # Write values to camera
                    try:
                        print('Write to camera')
                        print('CenterX: ' + str(center[0]))
                        print('CenterY: ' + str(center[1]))
                        cam.LensShadingCenterOffsetX.set(center[0])
                        cam.LensShadingCenterOffsetY.set(center[1])
                        print(correctionValuesSmooth)
                        for index in range(len(correctionValuesSmooth)):
                            cam.LensShadingIndex.set(index)
                            cam.LensShadingValue.set(float(correctionValuesSmooth[index]))
                        #cam.LensShadingValueAll.set(bytes(integers_to_bytes(lensCorrectionDec, 2)))
                        #cam.LensShadingEnable.set(True)
                    except (AttributeError, VmbFeatureError):
                        print("Error")
                        pass
                    exit()

                # Debugging and evaluation
                if mode == '/e':
                    # cv2.circle(src, center, 550, (255, 0, 255), 3)

                    # Upscale and display ROI 
                    # SCALE_PERCENT = 20
                    # width = int(src.shape[1] * SCALE_PERCENT / 100)
                    # height = int(src.shape[0] * SCALE_PERCENT / 100)
                    # dim = (width, height)
                    # src = cv2.resize(src, dim, interpolation = cv2.INTER_AREA)
                    # detectedThres = cv2.resize(detectedThres, dim, interpolation = cv2.INTER_AREA)

                    # cv2.imshow("detected", detectedThres)
                    # cv2.imshow("src", src)

                    ax.clear()
                    ax.plot(correctionValues)
                    ax.plot(correctionValuesSmooth)

                    plt.pause(0.05)


if __name__ == '__main__':
    main()


